// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:wyatt_ui_components/wyatt_ui_components.dart';

/// A registry for components of type [T], providing easy access and management.
class ComponentRegistry<T extends Component> {
  /// Constructs a [ComponentRegistry] with the given [components].
  const ComponentRegistry({required Map<String, T> components})
      : _components = components;

  /// Factory constructor to create a [ComponentRegistry]
  /// containing only a single [component].
  factory ComponentRegistry.solo(T component) => ComponentRegistry(
        components: {
          'solo': component,
        },
      );

  final Map<String, T> _components;

  /// Returns a component by its [id], or the first
  /// component if [id] is null or not found.
  T? call([String? id]) => getComponent(id) ?? getFirstComponent();

  /// Returns a component by its [id], or null if not found.
  T? getComponent(String? id) => _components[id];

  /// Returns the first component in the registry, or null
  /// if the registry is empty.
  T? getFirstComponent() => _components.values.firstOrNull;

  /// Checks if the registry contains a component with the given [id].
  bool containsComponent(String id) => _components.containsKey(id);

  /// Returns the count of components in the registry.
  int get count => _components.length;

  /// Returns a list of all IDs of components in the registry.
  List<String> get ids => _components.keys.toList();

  /// Returns a string representation of the registry,
  /// showing the count and IDs.
  @override
  String toString() => '$count - $ids';
}
